from .plots import (
    forecast,
    forecast_stats,
    forecast_ensembles,
    forecast_records,
    retrospective,
    daily_averages,
    monthly_averages,
    annual_averages,
    flow_duration_curve,
    corrected_retrospective,
    corrected_month_average,
    corrected_day_average,
    corrected_scatterplots,

    plotly_figure_to_html_plot as plotly_figure_to_html,
)

__all__ = [
    'forecast',
    'forecast_stats',
    'forecast_ensembles',
    'forecast_records',
    'retrospective',
    'daily_averages',
    'monthly_averages',
    'annual_averages',
    'flow_duration_curve',
    'corrected_retrospective',
    'corrected_month_average',
    'corrected_day_average',
    'corrected_scatterplots',
]
